package com.zauzoo.gpsservice.controller;

import com.zauzoo.gpsservice.render.TileMapComponent;
import com.zauzoo.gpsservice.repository.HistoryRepository;
import org.apache.commons.compress.utils.IOUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.PropertySource;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;

import java.io.FileInputStream;
import java.io.IOException;

/**
 * Kontroler pro mapové renderování.
 * @author Ondřej Chrbolka
 * @version 1.0
 *
 * */
@Controller
@RequestMapping("map")
@PropertySource("conf.properties")
public class MapController {
    /**
     * Komponenta pro vykreslování mapových dlaždic.
     * */
    private final TileMapComponent tileMapComponent;

    /**
     * Historie automobilů.
     * */
    private final HistoryRepository historyRepository;

    @Autowired
    public MapController(TileMapComponent tileMapComponent, HistoryRepository historyRepository) {
        this.tileMapComponent = tileMapComponent;
        this.historyRepository = historyRepository;
    }

    /**
     * Požadavek na vyrenderování mapy s bodem.
     *
     * @param latitude zeměpisná délka.
     * @param longitude zeměpisná šířka.
     * @param zoom úroveň zoomu.
     *
     * @return Mapový podklad.
     * */
    @RequestMapping(value = "tile", method = RequestMethod.GET, produces = MediaType.IMAGE_PNG_VALUE)
    public @ResponseBody byte[] getAddress(@RequestParam("latitude") Double latitude, @RequestParam("longitude") Double longitude,
                                           @RequestParam(value = "zoom", defaultValue = "10") Integer zoom) throws IOException {
        return IOUtils.toByteArray(new FileInputStream(tileMapComponent.mapRender(latitude, longitude, zoom).toFile()));
    }

    /**
     * Požadavek na vyrenderování mapy s bodem.
     *
     * @param latitude zeměpisná délka.
     * @param longitude zeměpisná šířka.
     * @param zoom úroveň zoomu.
     * @param width šířka obrázku v mapových dlažidích.
     * @param height výška obrázku v mapových dlažidích.
     *
     * @return Mapový podklad.
     * */
    @RequestMapping(value = "tile/detail", method = RequestMethod.GET, produces = MediaType.IMAGE_PNG_VALUE)
    public @ResponseBody Object getAddress(@RequestParam("latitude") Double latitude, @RequestParam("longitude") Double longitude,
                                      @RequestParam(value = "zoom", defaultValue = "10") Integer zoom,
                                      @RequestParam(value = "width") Integer width, @RequestParam(value = "height") Integer height) throws IOException {

        try{
            return IOUtils.toByteArray(new FileInputStream(tileMapComponent.mapRender(latitude, longitude, zoom, width, height).toFile()));
        } catch (ArrayIndexOutOfBoundsException e){
            return ResponseEntity.status(HttpStatus.BAD_REQUEST).body(null);
        }
    }
}
