package com.zauzoo.gpsservice.entity;

import org.openstreetmap.osmosis.core.domain.v0_6.Node;
import org.openstreetmap.osmosis.core.domain.v0_6.Tag;
import org.openstreetmap.osmosis.core.domain.v0_6.Way;

import javax.persistence.Transient;
import javax.persistence.Id;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import java.io.Serializable;
import java.util.*;

/**
 * Reprezentace mapové oblasti.
 * @author Ondřej Chrbolka
 * @version 1.0
 *
 * */
@Entity
public class Tie implements Serializable {
    /**
     * Status mapového podkladu, reprezentuje stav mapy ve kterém se nachází.
     * */
    public enum STATUS{
        /**
         * Mapový podklad je stahován z API.
         * */
        DOWNLOADING(false),
        /**
         * Mapový podklad se zpracovává pro další použití.
         * */
        PROCESSING(false),
        /**
         * Mapový podklad je importován do databáze.
         * */
        IMPORTING(true),
        /**
         * Mapový podklad je nepoužitelný.
         * */
        FAILED(false),
        /**
         * Mapový podklad je v pořádku.
         * */
        COMPLETE(false),
        /**
         * Mapový podklad bude v budoucnu nahrazen nahrazen větší oblastí.
         * */
        SLAVER(true);

        /**
         * Indikace zda-li je mapový podklad použitelný pro zpracování
         * */
        private boolean ussable;
        /**
         * Inicializace.
         * @param ussable Indikace zda-li je mapový podklad použitelný pro zpracování
         * */
        STATUS(boolean ussable) {
            this.ussable = ussable;
        }
        /**
         * Vrátí status mapového objektu.
         * */
        public boolean isUssable() {
            return ussable;
        }
    }

    /**
     * Primární klíč
     * */
    @Id
    @GeneratedValue
    private Long Id;

    /**
     * Status ve kterém se mapový podklad nachází.
     * */
    @Enumerated(EnumType.STRING)
    private STATUS status;

    /**
     * Cesty v mapovém podkladu.
     * */
    @Transient
    private Map<Long, Way> ways = new HashMap<>();

    /**
     * Body v mapovém podkladu.
     * */
    @Transient
    private Map<Long, Node> nodes = new HashMap<>();

    /**
     * Nody označené jako domy.
     * */
    @Transient
    private Map<String, List<Node>> houses = new HashMap<>();

    /**
     * Nody označené jako domy.
     * */
    @Transient
    private List<Tag> relationsNames = new ArrayList<>();

    /**
     * Ulice.
     * */
    @Transient
    private Map<String, List<Way>> streets = new HashMap<>();

    /**
     * Čas vytvoření
     * */
    private Date dateOfCreation;

    /**
     * Cesta k souboru s mapovým podkladem
     * */
    private String file;

    /**
     * Severní hraniční bod mapového podkladu
     * */
    private Double north;

    /**
     * Východní hraniční bod mapového podkladu
     * */
    private Double east;

    /**
     * Západní hraniční bod mapového podkladu
     * */
    private Double west;

    /**
     * Jižní hraniční bod mapového podkladu
     * */
    private Double south;
    /**
     * Adresa sektoru pro caschování X.
     * */
    private Integer sectorx;
    /**
     * Adresa sektoru pro caschování Y.
     * */
    private Integer sectory;

    /**
     * Inicializace.
     *
     * @param status Status ve kterém se mapový podklad nachází.
     *
     * */
    public Tie(STATUS status) {
        this.status = status;
    }

    /**
     * Inicializace.
     *
     * */
    public Tie(STATUS status, Double north, Double east, Double west, Double south) {
        this.status = status;
        this.north = north;
        this.east = east;
        this.west = west;
        this.south = south;
    }

    /**
     * Inicializace.
     *
     * */
    public Tie(STATUS status, Double north, Double east, Double west, Double south, Integer sectorx, Integer sectory) {
        this.status = status;
        this.north = north;
        this.east = east;
        this.west = west;
        this.south = south;
        this.sectorx = sectorx;
        this.sectory = sectory;
    }

    public Tie() {
    }

    public Tie(String file) {
        this.file = file;
    }

    public STATUS getStatus() {
        return status;
    }

    public void setStatus(STATUS status) {
        this.status = status;
    }

    public Map<Long, Way> getWays() {
        return ways;
    }

    public Map<Long, Node> getNodes() {
        return nodes;
    }

    public void setWays(Map<Long, Way> ways) {
        this.ways = ways;
    }

    public void setNodes(Map<Long, Node> nodes) {
        this.nodes = nodes;
    }

    public Long getId() {
        return Id;
    }

    public void setId(Long id) {
        this.Id = id;
    }

    public Date getDateOfCreation() {
        return dateOfCreation;
    }

    public void setDateOfCreation(Date dateOfCreation) {
        this.dateOfCreation = dateOfCreation;
    }

    public String getFile() {
        return file;
    }

    public void setFile(String file) {
        this.file = file;
    }

    public Map<String, List<Node>> getHouses() {
        return houses;
    }

    public void setHouses(Map<String, List<Node>> houses) {
        this.houses = houses;
    }

    public Map<String, List<Way>> getStreets() {
        return streets;
    }

    public void setStreets(Map<String, List<Way>> streets) {
        this.streets = streets;
    }

    public Double getNorth() {
        return north;
    }

    public void setNorth(Double north) {
        this.north = north;
    }

    public Double getEast() {
        return east;
    }

    public void setEast(Double east) {
        this.east = east;
    }

    public Double getWest() {
        return west;
    }

    public void setWest(Double west) {
        this.west = west;
    }

    public Double getSouth() {
        return south;
    }

    public void setSouth(Double south) {
        this.south = south;
    }

    public Integer getSectorx() {
        return sectorx;
    }

    public void setSectorx(Integer sectorx) {
        this.sectorx = sectorx;
    }

    public Integer getSectory() {
        return sectory;
    }

    public void setSectory(Integer sectory) {
        this.sectory = sectory;
    }

    public List<Tag> getRelationsNames() {
        return relationsNames;
    }

    public void setRelationsNames(List<Tag> relationsNames) {
        this.relationsNames = relationsNames;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Tie tie = (Tie) o;
        return Objects.equals(north, tie.north) &&
                Objects.equals(east, tie.east) &&
                Objects.equals(west, tie.west) &&
                Objects.equals(south, tie.south);
    }

    @Override
    public int hashCode() {
        return Objects.hash(north, east, west, south);
    }
}
