package com.zauzoo.gpsservice.maps;

import com.zauzoo.gpsservice.entity.Address;
import com.zauzoo.gpsservice.entity.AddressBuilder;
import com.zauzoo.gpsservice.entity.Tie;
import org.openstreetmap.osmosis.core.domain.v0_6.Node;
import org.openstreetmap.osmosis.core.domain.v0_6.Tag;
import org.openstreetmap.osmosis.core.domain.v0_6.Way;
import org.openstreetmap.osmosis.core.domain.v0_6.WayNode;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Komponenta pro geografické a matematické výpočty.
 * @author Ondřej Chrbolka
 * @version 1.0
 *
 * */
@Component
public class AddressTriangulation {
    /**
     * Poloměr zěmě.
     * */
    private static final Double EARTH_RADIUS = 637e3;

    /**
     * Výchozí rádius pokud není k dispozici žádný jiný.
     * */
    public static final Double DEFAULT_RADIUS = 500.0;
    public static final String DEFAULT_RADIUS_S = "500.0";

    /**
     * Určení státu.
     * */
    private final StateTriangulation stateTriangulation;

    @Autowired
    public AddressTriangulation(StateTriangulation stateTriangulation) {
        this.stateTriangulation = stateTriangulation;
    }

    /**
     * Najde v zadaném mapovém podkladu přibližnou adresu na základě výchozího bodu.
     *
     * @param tie mapový podklad.
     * @param latitude zeměpisná délka výchozího bodu.
     * @param longitude zeměpisná šířka výchozího bodu.
     *
     * */
    public Address findPossibleAddress(final Tie tie, final Double latitude, final Double longitude){
        return findPossibleAddress(tie, latitude, longitude, DEFAULT_RADIUS);
    }

    /**
     * Najde v zadaném mapovém podkladu přibližnou adresu na základě výchozího bodu.
     *
     * @param tie mapový podklad.
     * @param latitude zeměpisná délka výchozího bodu.
     * @param longitude zeměpisná šířka výchozího bodu.
     * @param radius poloměr tolerance.
     *
     * */
    public Address findPossibleAddress(Tie tie, Double latitude, Double longitude, Double radius) {
        final List<PossibleAddress> possibleWays = new ArrayList<>();
        final List<PossibleHouse> possibleHouses = new ArrayList<>();

        tie.getWays().forEach((id, way) -> {
            if(way.getWayNodes().size() >= 2){
                Node nodePrev, nodeNext;
                List<WayNode> nodes = way.getWayNodes();

                for (int i = 1; i < nodes.size(); ++i){
                    nodePrev = tie.getNodes().get(nodes.get(i - 1).getNodeId());
                    nodeNext = tie.getNodes().get(nodes.get(i).getNodeId());

                    BigDecimal length = checkCrossing(nodePrev, nodeNext, latitude, longitude, radius);
                    PossibleAddress address = new PossibleAddress(length, way);
                    address.setStreet(getAddressFromTag(way));
                    possibleWays.add(address);

                    if(length.compareTo(new BigDecimal(0)) >= 0)
                        address.setInside(true);
                     else
                        address.setInside(false);

                }
            }
        });
        possibleWays.sort(Comparator.comparing(o -> o.length));

        if(possibleWays.isEmpty()){
            String street = "";

            if(tie.getHouses().containsKey(street)){
                for (Node node : tie.getHouses().get(street)){
                    Double length = calculateDistance(latitude, node.getLatitude(), longitude, node.getLongitude());
                    PossibleHouse newHouse = new PossibleHouse(length, node);

                    if(isInside(latitude, longitude, radius, node.getLatitude(), node.getLongitude()))
                        newHouse.setInside(true);
                    else
                        newHouse.setInside(false);

                    possibleHouses.add(newHouse);
                }
            }
        } else {
            possibleWays.stream().forEach(possibleAddress -> {
                String street = getAddressFromTag(possibleAddress.getWay());
                if(tie.getHouses().containsKey(street)){
                    for (Node node : tie.getHouses().get(street)){
                        Double length = calculateDistance(latitude, node.getLatitude(), longitude, node.getLongitude());
                        PossibleHouse newHouse = new PossibleHouse(length, node);

                        if(isInside(latitude, longitude, radius, node.getLatitude(), node.getLongitude()))
                            newHouse.setInside(true);
                        else
                            newHouse.setInside(false);

                        possibleHouses.add(newHouse);
                    }
                }
            });
        }

        possibleHouses.sort(Comparator.comparingDouble(o -> o.length));

        List<PossibleHouse> withRadiusH = possibleHouses.stream().filter(PossibleHouse::isInside).collect(Collectors.toList());
        List<PossibleAddress> withRadiusW = possibleWays.stream().filter(PossibleAddress::isInside).collect(Collectors.toList());

        String state = "";
        try {
            state = stateTriangulation.getState(latitude, longitude);
        } catch (Exception e) {

        }

        String possibleStreet = getAddressFromTag(possibleHouses.get(0).getNode());

        if(possibleStreet.isEmpty())
            if(withRadiusW.isEmpty())
                possibleStreet = "";
            else
                possibleStreet = getAddressFromTag(withRadiusW.get(0).getWay());



        Address a = AddressBuilder.anAddress()
                .withStreet(possibleStreet)
                .withNumber((withRadiusH.isEmpty() ? "" : getNumberFromTag(withRadiusH.get(0).getNode())) )
                .withCity(getCityFromTag(possibleWays, possibleHouses, tie.getRelationsNames()))
                .withPlace(getCustomFromTag(possibleWays, possibleHouses, "addr:place"))
                .withZip(getZipFromTag(possibleWays, possibleHouses, tie.getRelationsNames()))
                .withState(state).build();

        if(a.getCity().isEmpty() && !a.getPlace().isEmpty())
            a.setCity(a.getPlace());

        return a;
    }

    /**
     * Zkontroluje zdali se bod nachází v oblasti kruhu.
     * @param circle_x x kruhu.
     * @param circle_y y kruhu.
     * @param rad radius kruhu.
     * @param x bodu.
     * @param y bodu.
     *
     * @return true - nachází, false - je mimo.
     * */
    private boolean isInside(Double circle_x, Double circle_y, Double rad, Double x, Double y) {
        if ((x - circle_x) * (x - circle_x) +
                (y - circle_y) * (y - circle_y) <= rad * rad)
            return true;
        else
            return false;
    }

    /**
     * Provede separaci.
     * @param way cesta
     *
     * */
    private String getAddressFromTag(final Way way){
        for (Tag tag : way.getTags()) {
            if (tag.getKey().endsWith("addr:street") || tag.getKey().equals("name"))
                return tag.getValue();
        }
        return "";
    }

    /**
     * Provede separaci.
     * @param node bod.
     *
     * */
    private String getAddressFromTag(final Node node){
        for (Tag tag : node.getTags()) {
            if (tag.getKey().endsWith("addr:street") || tag.getKey().equals("name"))
                return tag.getValue();
        }
        return "";
    }

    /**
     * Provede separaci.
     * @param node bod
     *
     * */
    private String getNumberFromTag(final Node node){
        for (Tag tag : node.getTags()) {
            if (tag.getKey().endsWith("addr:housenumber"))
                return tag.getValue();
        }
        return "";
    }

    /**
     * Provede separaci.
     * @param ways cesty.
     * @param houses domy.
     *
     * @param relationsNames*/
    private String getCityFromTag(final List<PossibleAddress> ways, final List<PossibleHouse> houses, List<Tag> relationsNames){
        for (PossibleAddress way : ways){
            for (Tag tag : way.getWay().getTags()) {
                if (tag.getKey().endsWith("addr:city"))
                    return tag.getValue();
            }
        }

        for (PossibleHouse house : houses){
            for (Tag tag : house.getNode().getTags()) {
                if (tag.getKey().endsWith("addr:city"))
                    return tag.getValue();
            }
        }

        return "";
    }

    /**
     * Vrátí libovolný tag z kolekcí.
     * @param ways cesty.
     * @param houses domy.
     * @param name název klíče tagu.
     * @return hodnota tagu.
     * */
    private String getCustomFromTag(final List<PossibleAddress> ways, final List<PossibleHouse> houses, final String name){
        for (PossibleAddress address : ways){
            for (Tag tag : address.getWay().getTags()){
                if(tag.getKey().equals(name))
                    return tag.getValue();
            }
        }

        for (PossibleHouse house : houses){
            for (Tag tag : house.getNode().getTags())
                if(tag.getKey().equals(name))
                    return tag.getValue();
        }

        return "";
    }

    /**
     * Provede separaci.
     * @param ways cesty.
     * @param houses domy.
     *
     **/
    private String getZipFromTag(final List<PossibleAddress> ways, final List<PossibleHouse> houses, List<Tag> relationsNames){
        for (PossibleAddress way : ways){
            for (Tag tag : way.getWay().getTags()) {
                if (tag.getKey().endsWith("addr:postcode"))
                    return tag.getValue();
            }
        }

        for (PossibleHouse house : houses){
            for (Tag tag : house.getNode().getTags()) {
                if (tag.getKey().endsWith("addr:postcode"))
                    return tag.getValue();
            }
        }

        return "";
    }

    /**
     * Vypočítá vzdálenost se započítáním zakřivení země.
     *
     * @param lat1 zeměpisná délka 1.
     * @param lat2 zeměpisná délka 2.
     * @param lon1 zeměpisná šířka 1.
     * @param lon2 zeměpisná šířka 2.
     *
     * @return vzdálenost dvou bodů.
     * */
    protected Double calculateDistance(Double lat1, Double lat2, Double lon1, Double lon2){
        double l1 = lat1 * Math.PI/180;
        double l2 = lat2 * Math.PI/180;

        double deltaLat = (lat2 - lat1) * Math.PI/180;
        double deltaLon = (lon2 - lon1) * Math.PI/180;

        double a = Math.sin(deltaLat/2) * Math.sin(deltaLat/2) +
                   Math.sin(deltaLon/2) * Math.sin(deltaLon/2) *
                   Math.cos(l1) * Math.cos(l2);
        double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));

        return EARTH_RADIUS * c;
    }

    /**
     * Vypočítá vzdálenost bodu od přímky.
     *
     * @param node1 počíteční bod přímky.
     * @param node2 koncový bod přímky.
     * @param latitude zeměpisná délka bodu.
     * @param longitude zeměpisná šířka bodu.
     * @param radius radius tolerance.
     *
     * @return vzdálenost. Pokud je záporná je bod mimo radius.
     * */
    protected BigDecimal checkCrossing(Node node1, Node node2, final Double latitude, final Double longitude, final Double radius) {
        double longitudeA = node1.getLongitude();
        double latitudeA = node1.getLatitude();

        double longitudeB = node2.getLongitude();
        double latitudeB = node2.getLatitude();

        List<Double> linePoints = calculateTheLine(latitudeA, latitudeB, longitudeA, longitudeB);

        GFG.Pair A = new GFG.Pair(new BigDecimal(longitudeA), new BigDecimal(latitudeA));
        GFG.Pair B = new GFG.Pair(new BigDecimal(longitudeB), new BigDecimal(latitudeB));
        GFG.Pair E = new GFG.Pair(new BigDecimal(longitude),new BigDecimal( latitude));

        if(checkCollision(linePoints.get(0), linePoints.get(1), linePoints.get(2), latitude, longitude, radius)){
            return GFG.minDistance(A, B, E);
        } else {
            return GFG.minDistance(A, B, E).multiply(new BigDecimal(-1.0));
        }
    }

    /**
     * Spočítá parametry pro rovnici přímky.
     *
     * @param x1 Počáteční x souřadnice.
     * @param x2 Koncová x souřadnice.
     * @param y1 Počáteční y souřadnice.
     * @param y2 Koncová y souřadnice.
     *
     * @return parametry rovnice přímky.
     *
     * */
    public List<Double> calculateTheLine(final double x1, final double x2, final double y1, final double y2){
        double v1 = x2 - x1;
        double v2 = y2 - y1;

        if(v2 < 0)
            v2 *= -1;
        else
            v1 *= -1;

        double[] nv = {v2 ,v1};

        double a = nv[0];
        double b = nv[1];
        double c = ((nv[0] * x1) + (nv[1] * y1)) * -1;

        return Arrays.asList(a, b, c);
    }

    /**
     * Detekuje kolizi s cestou.
     *
     * @param a konstanta obecné rovnice přímky
     * @param b konstanta obecné rovnice přímky
     * @param c konstanta obecné rovnice přímky
     * @param x souřadnice bodu
     * @param y  souřadnice bodu
     * @param radius poloměr tolerance
     *
     * @return true - kolize false - bez kolize
     * */
    public boolean checkCollision(Double a, Double b, Double c, Double x, Double y, Double radius) {
        double dist = (Math.abs(a * x + b * y + c)) /
                Math.sqrt(a * a + b * b);

        if (radius >= dist)
            return true;
        else
            return false;
    }

    /**
     * Datová třída pro možného kandidáta na adresu.
     *
     * @author Ondřej Chrbolka
     * @version 1.0
     * */
    private class PossibleAddress {
        /**
         * Vzdálenost.
         * */
        private BigDecimal length;
        /**
         * Ulice.
         * */
        private String street;
        /**
         * Cesta.
         * */
        private Way way;

        /**
         * Rozpoznání zdali je uvnitř radiusu.
         * */
        private boolean inside;


        public PossibleAddress(BigDecimal length, Way way) {
            this.length = length;
            this.way = way;
        }

        public BigDecimal getLength() {
            return length;
        }

        public void setLength(BigDecimal length) {
            this.length = length;
        }

        public String getStreet() {
            return street;
        }

        public void setStreet(String street) {
            this.street = street;
        }

        public Way getWay() {
            return way;
        }

        public void setWay(Way way) {
            this.way = way;
        }

        public boolean isInside() {
            return inside;
        }

        public void setInside(boolean inside) {
            this.inside = inside;
        }
    }

    /**
     * Datová třída pro možného kandidáta na číslo domu.
     *
     * @author Ondřej Chrbolka
O     * @version 1.0
     * */
    private class PossibleHouse{
        /**
         * Vzdálenost.
         * */
        private Double length;
        /**
         * Bod.
         * */
        private Node node;

        /**
         * Rozpoznání zdali je uvnitř radiusu.
         * */
        private boolean inside;

        public PossibleHouse(Double length, Node node) {
            this.length = length;
            this.node = node;
        }

        public Double getLength() {
            return length;
        }

        public void setLength(Double length) {
            this.length = length;
        }

        public Node getNode() {
            return node;
        }

        public void setNode(Node way) {
            this.node = way;
        }

        public boolean isInside() {
            return inside;
        }

        public void setInside(boolean inside) {
            this.inside = inside;
        }
    }
}
