package com.zauzoo.gpsservice.maps;

import com.zauzoo.gpsservice.entity.Tie;

import org.openstreetmap.osmosis.core.container.v0_6.EntityContainer;
import org.openstreetmap.osmosis.core.container.v0_6.NodeContainer;
import org.openstreetmap.osmosis.core.container.v0_6.RelationContainer;
import org.openstreetmap.osmosis.core.container.v0_6.WayContainer;
import org.openstreetmap.osmosis.core.domain.v0_6.*;
import org.openstreetmap.osmosis.core.task.v0_6.Sink;
import org.openstreetmap.osmosis.xml.common.CompressionMethod;

import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Component;

import javax.validation.constraints.NotNull;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Třída umožňující stažení a další zpracování OSM mapových podkladů.
 * @author Ondřej Chrbolka
 *
 * @version 1.0
 **/
@Component
@PropertySource("conf.properties")
public class OsmMapComponent implements MapComponent {
    /**
     * Proměnné prostředí
     * */
    protected final Environment environment;

    /**
     * API URL
     * */
    private String apiURL;

    /**
     * Inicializace
     *
     * */

    public OsmMapComponent(Environment environment) {
        this.environment = environment;
        this.apiURL = environment.getProperty("openstreetmap.source.server.main");
    }

    /**
     * Stáhne osm soubor s mapovou oblastí do určené lokace.
     *
     * @param location Umístění na serveru, kde bude soubor uchován k dalšímu zpracování.
     * @throws IOException Vstupně výstupní chyba stahování.
     *
     * */
    public void download(@NotNull final Path location,@NotNull final String urlToDownload) throws IOException {
        URL url = new URL(urlToDownload);
        ReadableByteChannel channel = Channels.newChannel(url.openStream());

        FileOutputStream outputStream = new FileOutputStream(location.toFile());
        FileChannel fchannel = outputStream.getChannel();

        outputStream.getChannel().transferFrom(channel, 0, Long.MAX_VALUE);
    }

    @Override
    public Tie download(@NotNull final Double longitude, @NotNull final Double latitude, @NotNull String location) throws IOException {
        Tie newTie = new Tie(Tie.STATUS.DOWNLOADING);

        String apiUrl = createAPIURL(longitude, latitude, newTie);
        location = String.format("%s%d.osm", location, Math.abs(newTie.hashCode()));
        newTie.setFile(location);

        download(Paths.get(location), apiUrl);

        return newTie;
    }

    @Override
    public Tie processOSM(@NotNull final Path location, @NotNull final Tie tie) throws FileNotFoundException {
        tie.setStatus(Tie.STATUS.PROCESSING);
        FileInputStream osmInput = new FileInputStream(location.toFile());
        ServerXMLReader osmosisReader = new ServerXMLReader(location.toFile(), true, CompressionMethod.None);

        SpringSink sink = new SpringSink(osmInput) {
            @Override
            public void complete() {
                this.close();
            }
        };
        Map<Long, Way> ways = osmosisReader.read(sink);

        tie.setWays(ways);
        tie.setNodes(sink.getNodes());
        tie.setHouses(sink.getHouses());
        tie.setStreets(sink.getStreets());
        tie.setRelationsNames(sink.getRelationsNames());

        tie.setStatus(Tie.STATUS.COMPLETE);

        return tie;
    }

    /**
     * Vytvoří URL s požadavkem na OpenStreetMap server.
     * @param longtitude zeměpisná šířka
     * @param latitude zeměpisná délka
     *
     * @return URL adresa s požadavkem na server
     *
     * */
    protected String createAPIURL(@NotNull final Double longtitude, @NotNull final Double latitude, @NotNull final Tie tie){
        double toleration = Double.parseDouble(environment.getProperty("openstreetmap.source.side.toleration"));

        tie.setWest(longtitude - toleration);
        tie.setEast(longtitude + toleration);
        tie.setSouth(latitude + toleration);
        tie.setNorth(latitude - toleration);

        return createURL(tie);
    }

    /**
     * Vytvoří URL pro API.
     * @param tie Mapový podklad s informacema.
     *
     * @return URL pro API.
     * */
    protected final String createURL(@NotNull Tie tie) {
        String west = String.valueOf(tie.getWest()).replaceAll(",", ".");
        String east = String.valueOf(tie.getEast()).replaceAll(",", ".");
        String north = String.valueOf(tie.getNorth()).replaceAll(",", ".");
        String south = String.valueOf(tie.getSouth()).replaceAll(",", ".");

        return String.format(this.apiURL, west, north, east, south);
    }

    public String getApiURL() {
        return apiURL;
    }

    public void setApiURL(String apiURL) {
        this.apiURL = apiURL;
    }

    /**
     * Vnitřní třída reprezentující callbak při zpracovaní OSM map.
     *
     * @author Ondřej Chrbolka
     * @version 1.0
     * */
    protected abstract class SpringSink implements Sink {
        /**
         * Ulice.
         * */
        private final Map<String, List<Way>> streets = new HashMap<>();

        /**
         * Cesty nalezené v OSM souboru.
         * */
        private final Map<Long, Way> ways = new HashMap<>();

        /**
         * Body nalezené v mapovém podkladu
         * */
        private final Map<Long, Node> nodes = new HashMap<>();

        /**
         * Nody označené jako domy.
         * */
        private Map<String , List<Node>> houses = new HashMap<>();

        /**
         * Vstupní soubor
         * */
        private final FileInputStream fileInputStream;
        /**
         * Jména vztahů
         * */
        private List<Tag> relationsNames;

        /**
         * Inicializuje třídu.
         * @param fileInputStream Vstupní stream souboru, který bude po dokončení čtení zavřen.
         * */
        public SpringSink(FileInputStream fileInputStream) {
            Objects.requireNonNull(fileInputStream);
            this.fileInputStream = fileInputStream;
        }

        @Override
        public void process(EntityContainer entityContainer) {
            if(entityContainer instanceof WayContainer) {
                Way way = ((WayContainer) entityContainer).getEntity();
                Collection<Tag> tags = way.getTags();
                final boolean[] highway = {false};
                tags.forEach( tag -> {
                    if(tag.getKey().equals("highway"))
                        highway[0] = true;

                    if ((tag.getKey().equals("name") && highway[0]) || tag.getKey().endsWith("addr:street")){
                        ways.put(way.getId(), way);
                        String streetKey = tag.getValue();

                        if(streets.containsKey(streetKey)){
                            streets.get(streetKey).add(way);
                        } else {
                            streets.put(streetKey, new ArrayList<>());
                            streets.get(streetKey).add(way);
                        }
                    }
                });
            }
            if(entityContainer instanceof NodeContainer){
                Node node = ((NodeContainer)entityContainer).getEntity();

                if(node.getTags().size() >= 2){
                    Tag address = null, housenumber = null;

                    for (Tag tag : node.getTags()){
                        if(tag.getKey().equals("addr:housenumber")) housenumber = tag;
                        if(tag.getKey().equals("addr:street")) address = tag;
                    }

                    if(housenumber != null){
                        String addressKey = address == null ? "" : address.getValue();

                        if(!houses.containsKey(addressKey)){
                            houses.put(addressKey, new ArrayList<>());
                            houses.get(addressKey).add(node);
                        } else {
                            List<Node> list = houses.get(addressKey);
                            list.add(node);
                        }
                    }
                }
                nodes.put(node.getId(), node);
            }

            if(entityContainer instanceof RelationContainer){
                Relation relation = ((RelationContainer)entityContainer).getEntity();
                Collection<Tag> tags = relation.getTags();

                if(tags.size() > 0){
                    List<Tag> names = tags.stream().filter(tag -> tag.getKey().endsWith("name")).collect(Collectors.toList());
                    this.relationsNames = names;
                }
            }
        }

        @Override
        public void close() {
            try {
                fileInputStream.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        @Override
        public void initialize(Map<String, Object> metaData) { }

        /**
         * Vrátí cesty.
         *
         * @return List cest.
         * */
        public Map<Long, Way> getWays() {
            return ways;
        }

        /**
         * Vrátí body.
         *
         * @return Mapa bodů
         * */
        public Map<Long, Node> getNodes() {
            return nodes;
        }

        public Map<String, List<Way>> getStreets() {
            return streets;
        }

        public Map<String, List<Node>> getHouses() {
            return houses;
        }

        public FileInputStream getFileInputStream() {
            return fileInputStream;
        }

        public void setHouses(Map<String, List<Node>> houses) {
            this.houses = houses;
        }

        public List<Tag> getRelationsNames() {
            return relationsNames;
        }

        public void setRelationsNames(List<Tag> relationsNames) {
            this.relationsNames = relationsNames;
        }

    }
}
