package com.zauzoo.gpsservice.maps;

import com.zauzoo.gpsservice.core.ConfigurationChecker;
import com.zauzoo.gpsservice.entity.Tie;
import com.zauzoo.gpsservice.repository.TieRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import javax.validation.constraints.NotNull;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Objects;
import java.util.Queue;

/**
 * Komponenta pro mapové caschování.
 * @author Ondřej Chrbolka
 * @version 1.0
 *
 * */
@Component
@PropertySource("conf.properties")
public class SectorMapComponent extends OsmMapComponent {
    /**
     * Cashovací mapa.
     * */
    protected final HashMap<String, Tie> cachedMap = new HashMap<>();

    /**
     * Proměnné prostředí.
     **/
    protected Environment environment;

    /**
     * Fronta pro stánutí nových mapových podkladů
     * */
    protected final Queue<Tie> toDownload = new LinkedList<>();

    /**
     * Repozitář mapových podkladů.
     * */
    protected final TieRepository tieRepository;

    /**
     * Inicializace
     */
    @Autowired
    public SectorMapComponent(Environment environment, TieRepository tieRepository) {
        super(environment);
        this.environment = environment;
        this.tieRepository = tieRepository;
    }

    @Override
    public Tie download(@NotNull Double longitude, @NotNull Double latitude, @NotNull String location) throws IOException {
        int sectorX = (int)(longitude / Double.valueOf(environment.getProperty("openstreetmap.source.side.toleration")));
        int sectorY = (int)(latitude / Double.valueOf(environment.getProperty("openstreetmap.source.side.toleration")));
        String coordinates = String.format("%d %d", sectorX, sectorY);

        Tie tie = (cachedMap.containsKey(coordinates)? cachedMap.get(coordinates) : new Tie(Tie.STATUS.DOWNLOADING));
        if(tie.getStatus() == Tie.STATUS.COMPLETE)
            return tie;

        String apiUrl = createAPIURL(sectorX, sectorY, tie);

        location = String.format("%s%d.osm", location, Math.abs(tie.hashCode()));
        tie.setFile(location);

        download(Paths.get(location), apiUrl);

        tie.setSectorx(sectorX);
        tie.setSectory(sectorY);

        cachedMap.put(coordinates, tie);

        return processOSM(Paths.get(tie.getFile()), tie);
    }

    /**
     * Vytvoří URL podle sektoru.
     *
     * @param sectorX souřadnice sektoru X.
     * @param sectorY souřadnice sektoru Y.
     * @param tie mapový podklad pro uložení.
     *
     * @return URL pro stažení mapového podkladu.
     *
     * */
    protected String createAPIURL(@NotNull Integer sectorX, @NotNull Integer sectorY, @NotNull Tie tie) {
        double toleration = Double.parseDouble(environment.getProperty("openstreetmap.source.side.toleration"));

        Double baseLatitude = sectorY * toleration;
        Double baseLongitude = sectorX * toleration;

        tie.setNorth(baseLatitude);
        tie.setSouth(baseLatitude + toleration);
        tie.setWest(baseLongitude);
        tie.setEast(baseLongitude + toleration);

        return createURL(tie);
    }

    /**
     * Stáhne mapovou dlažidici pro optimalizaci.
     * @param tie Mapová dlaždice.
     *
     * @return mapová dlaždice.
     * */
    public Tie download(Tie tie) throws IOException {
        String location = String.format("%s%d.osm", ConfigurationChecker.PATH, Math.abs(tie.hashCode()));
        tie.setFile(location);

        Path path = Paths.get(location);
        this.download(path, createURL(tie));

        return super.processOSM(path, tie);
    }

    @Override
    public Tie processOSM(@NotNull Path location, @NotNull Tie tie) throws FileNotFoundException {
        addToQueue(tie);

        Double next = Double.parseDouble(environment.getProperty("openstreetmap.source.side.toleration"));
        this.getToDownload().add(new Tie(Tie.STATUS.SLAVER, tie.getNorth() + next, tie.getEast(), tie.getWest(), tie.getSouth() + next, tie.getSectorx(), tie.getSectory() + 1));
        this.getToDownload().add(new Tie(Tie.STATUS.SLAVER, tie.getNorth() - next, tie.getEast(), tie.getWest(), tie.getSouth() - next, tie.getSectorx(), tie.getSectory() - 1));
        this.getToDownload().add(new Tie(Tie.STATUS.SLAVER, tie.getNorth(), tie.getEast() + next, tie.getWest() + next, tie.getSouth(), tie.getSectorx() + 1, tie.getSectory()));
        this.getToDownload().add(new Tie(Tie.STATUS.SLAVER, tie.getNorth(), tie.getEast() - next, tie.getWest() - next, tie.getSouth(), tie.getSectorx() - 1, tie.getSectory()));

        return super.processOSM(location, tie);
    }

    @Scheduled(fixedRate = 5000)
    public void downloadNew() throws IOException {
        if(!toDownload.isEmpty()){
            Tie tie = toDownload.remove();

            if(tieRepository.countBySectorxAndSectory(tie.getSectorx(), tie.getSectory()) == 0){
                this.download(tie);
                addToQueue(tie);
            }
        }
    }

    public Queue<Tie> getToDownload() {
        return toDownload;
    }

    /**
     * Přidá objekt do cashovací mapy.
     *
     * @param tie Mapový podklad.
     * */
    public void addToQueue(final Tie tie){
        Objects.requireNonNull(tie.getSectorx());
        Objects.requireNonNull(tie.getSectory());

        String coordinates = String.format("%d %d", tie.getSectorx(), tie.getSectory());

        cachedMap.put(coordinates, tie);
    }

    public HashMap<String, Tie> getCachedMap() {
        return cachedMap;
    }
}
