package com.zauzoo.gpsservice.maps;

import org.geotools.data.DataStore;
import org.geotools.data.DataStoreFinder;
import org.geotools.data.FeatureSource;
import org.geotools.factory.CommonFactoryFinder;
import org.geotools.feature.FeatureCollection;
import org.geotools.util.factory.GeoTools;
import org.jboss.logging.Logger;
import org.locationtech.jts.geom.Coordinate;
import org.locationtech.jts.geom.GeometryFactory;
import org.locationtech.jts.geom.Point;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.filter.Filter;
import org.opengis.filter.FilterFactory2;
import org.opengis.filter.expression.Expression;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Component;


import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Třída pro určení státu na základě předložených mapových podkladů.
 * @author Ondřej Chrbolka
 * @version 1.0
 *
 * */
@Component
@PropertySource("conf.properties")
public class StateTriangulation {
    private final Logger logger;

    /**
     * Proměnné prostředí.
     * */
    private final Environment environment;

    /**
     * Jmenovité atributy v souboru.
     * */
    private String typeName;

    /**
     * Úložiště dat.
     * */
    private DataStore dataStore;

    /**
     * Indikace fungování komponenty.
     * */
    private boolean ready = true;

    /**
     * Přípony potřebné pro získání informací o hranicích státu.
     * */
    private static final List<String> FILE_TYPES = Arrays.asList(".shp", ".dbf", ".prj", ".qix", ".shx");

    @Autowired
    public StateTriangulation(Environment environment) throws IOException {
        this.environment = environment;
        this.logger = Logger.getLogger(StateTriangulation.class);

        //Kontrola dostupnosti shape filu pro zjištění státu
        String shape_file = environment.getProperty("global.states.database.file");

        if(shape_file.isEmpty() || shape_file == null){
            logger.warn(String.format("Shape file s hranicemi států není dostupný. Identifikace území nebude fungovat."));
            ready = false;
            return;
        }

        FILE_TYPES.forEach(file -> {
            Path path = Paths.get(shape_file.replace(".shp", file));
            if(!Files.exists(path)){
                ready = false;
                logger.warn(String.format("(%s) Shape file s hranicemi států není dostupný. Identifikace území nebude fungovat.", file));
            }
        });

        if(ready){
            String filePath = environment.getProperty("global.states.database.file");
            File file = new File(filePath);

            Map<String, Object> map = new HashMap<>();
            map.put("url", file.toURI().toURL());

            this.dataStore = DataStoreFinder.getDataStore(map);
            this.typeName = dataStore.getTypeNames()[0];
        }
    }

    /**
     * Vrátí název státu na základě zadaných kordinatů.
     * @param lat zeměpisná délka.
     * @param lon zeměpisná šířka.
     *
     * @return Název státu.
     * */
    public String getState(final Double lat, final Double lon) throws Exception {
        if(ready){
            FeatureSource<SimpleFeatureType, SimpleFeature> source = dataStore.getFeatureSource(typeName);
            FilterFactory2 filterFactory = CommonFactoryFinder.getFilterFactory2(GeoTools.getDefaultHints());

            GeometryFactory fac = new GeometryFactory();
            Point p = fac.createPoint(new Coordinate(lon, lat));

            SimpleFeatureType schema = source.getSchema();
            Expression propertyName = filterFactory.property(schema.getGeometryDescriptor().getName());

            Filter filter = filterFactory.contains(propertyName, filterFactory.literal(p));

            FeatureCollection<SimpleFeatureType, SimpleFeature> sub = source.getFeatures().subCollection(filter);
            if(sub.features().hasNext())
                return sub.features().next().getAttribute("name").toString();
            else
                return "";
        } else {
            return "";
        }
    }
}