package com.zauzoo.gpsservice.render;

import com.zauzoo.gpsservice.entity.History;
import com.zauzoo.gpsservice.maps.MapImage;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Component;

import javax.imageio.ImageIO;
import javax.validation.constraints.NotNull;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.math.BigDecimal;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

/**
 * Komponenta určená pro vykresolání mapových dlaždic.
 *
 * @author Ondřej Chrbolka
 * @version 1.0
 *
 * */
@Component
@PropertySource("conf.properties")
public class TileMapComponent implements TileComponent {
    /**
     * Výchozí hodnota šířky dlaždic.
     */
    private static final Integer WIDTH = 256;

    /**
     * Výchozí hodnota výšky dlaždic.
     */
    private static final Integer HEIGHT = 256;

    /**
     * Výchozí velikost hashe.
     * */
    public static final Integer DEFAULT_HASH_LENGTH = 64;

    /**
     * Výchozí velikost značky.
     * */
    public static final Integer MARKER_WIDTH = 10;

    public static final Color MARKER_COLOR = Color.RED;

    /**
     * Proměnné prostředí.
     * */
    private final Environment environment;

    @Autowired
    public TileMapComponent(Environment environment) {
        this.environment = environment;
    }

    /**
     * Provede stažení jednoho dílku.
     *
     * @param historyList  historie automobilů.
     * @param zoom  zoom level.
     * @param g grafika na vykreslení.
     * @param tile mapová dlaždice.
     * */
    protected void downloadMapTile(@NotNull List<History> historyList, @NotNull Integer zoom, TileInfo tile, Graphics g) throws IOException {
        downloadMapTile(historyList, zoom, tile, g, tile.getxTile(), tile.getyTile());
    }

    /**
     * Provede stažení jednoho dílku.
     *
     * @param historyList  historie automobilů.
     * @param zoom  zoom level.
     * @param g grafika na vykreslení.
     * @param tile mapová dlaždice.
     * @param indexX X umístění dílku.
     * @param indexY Y umístění dílku.
     * */
    protected void downloadMapTile(@NotNull List<History> historyList, @NotNull Integer zoom, TileInfo tile, Graphics g, int indexX, int indexY) throws IOException {
        String path = String.format("%s%s%d-%d-%d.png.png", environment.getProperty("openstreetmap.tile.download.folder"), File.separator, zoom, indexX, indexY);
        Path location = Paths.get(path);

        String main = environment.getProperty("openstreetmap.tile.main");
        String url = String.format("%s%d/%d/%d.png", main, zoom, indexX, indexY);

        if(!Files.exists(location)) download(location, url);

        for (History history : historyList) {
            TileInfo info = createAPIURL(history.getLatitude(), history.getLongtitude(), zoom);
            if (indexX == info.getxTile() && indexY == info.getyTile()){
                drawMarkToImage(Paths.get(path), info.getxCord(), info.getyCord());
            }
        }

        BufferedImage bi = ImageIO.read(new File(path));
        g.drawImage(bi, (indexX - tile.getxTile()) * WIDTH, (indexY - tile.getyTile()) * WIDTH, null);
    }

    /**
     * Provede vykreslování mapy.
     *
     * @param longitude zeměpisná délka.
     * @param latitude zeměpisná šířka.
     * @param zoom zoom mapového podkladu.
     * @param width šířka obrázku v mapových dlažidích.
     * @param height výška obrázku v mapových dlažidích.
     * */
    public Path mapRender(@NotNull Double latitude, @NotNull final Double longitude, final Integer zoom, final Integer width, final Integer height) throws IOException {
        int max = (int) Math.pow(2, zoom);

        TileInfo info = createAPIURL(latitude, longitude, zoom);
        int xcor = info.getxTile() - width;
        int ycor = info.getyTile() - height;

        if(xcor < 0 || ycor < 0 || info.getxTile() + width > max || info.getyTile() + height > max)
            throw new ArrayIndexOutOfBoundsException();

        int realWidth = width * 2 + 1;
        int realHeight = height * 2 + 1;

        BufferedImage result = new BufferedImage(realWidth * WIDTH, realHeight * WIDTH, BufferedImage.TYPE_INT_RGB);
        Graphics g = result.getGraphics();

        downloadMapTile(zoom, info, xcor, ycor, realWidth, realHeight, g);

        String hash = RandomStringUtils.random(DEFAULT_HASH_LENGTH, true, true);
        File output = new File(String.format("%s%s%s.png", environment.getProperty("openstreetmap.tile.download.folder"), File.separator, hash));
        ImageIO.write(result,"png", output);

        return Paths.get(output.getAbsolutePath());
    }

    /**
     * Provede vykreslování mapy.
     *
     * @param zoom zoom.
     * @param info informace o dlaždici.
     * @param xcor xová souřadnice.
     * @param ycor yová souřadnice.
     * @param realWidth šířka obrázku.
     * @param realHeight délka obrázku.
     * @param g grafika pro vykreslování finálního obrázku.
     *
     * */
    protected void downloadMapTile(Integer zoom, TileInfo info, int xcor, int ycor, int realWidth, int realHeight, Graphics g) throws IOException {
        for (int i = 0; i < realWidth; ++i){
            for (int j = 0; j < realHeight; ++j){
                String path =  String.format("%s%s%d-%d-%d.png", environment.getProperty("openstreetmap.tile.download.folder"), File.separator, zoom, i + xcor, j + ycor);

                String main = environment.getProperty("openstreetmap.tile.main");
                String url = String.format("%s%d/%d/%d.png", main, zoom, i + xcor, j + ycor);

                Path location = Paths.get(path);

                if(!Files.exists(location)) download(location, url);

                if((info.getxTile() == i + xcor && info.getyTile() == j + ycor))
                    drawMarkToImage(Paths.get(path), info.getxCord(), info.getyCord());

                BufferedImage bi = ImageIO.read(new File(path));
                g.drawImage(bi, i * WIDTH, j * WIDTH, null);
            }
        }
    }

    /**
     * Vyrendruje mapu na daných souřadnicích.
     *
     * @param longitude zeměpisná délka.
     * @param latitude zeměpisná šířka.
     * @param zoom zoom mapového podkladu.
     *
     * @return cesta k mapovému podkladu.
     * */
    public Path mapRender(@NotNull Double latitude, @NotNull final Double longitude, final Integer zoom) {
        MapImage mapImage = new MapImage(latitude, longitude, zoom);

        String path =  String.format("%s%s%d.png", environment.getProperty("openstreetmap.tile.download.folder"), File.separator, Math.abs(mapImage.hashCode())) ;
        TileInfo info = createAPIURL(latitude, longitude, zoom);

        try {
            download(Paths.get(path), info.getUrl());
            drawMarkToImage(Paths.get(path), info.getxCord(), info.getyCord());
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }

        return Paths.get(path);
    }

    /**
     * Zakreslí bod do obrázku.
     *
     * @param path vstupní soubor.
     * @param xCord xová souřadnice.
     * @param yCord yová souřadnice.
     * */
    private void drawMarkToImage(Path path, Integer xCord, Integer yCord) throws IOException {
        BufferedImage image = ImageIO.read(path.toFile());

        Graphics2D g = (Graphics2D) image.getGraphics();

        g.setComposite(AlphaComposite.Src);
        g.setPaint(MARKER_COLOR);
        g.setColor(MARKER_COLOR);
        g.setBackground(MARKER_COLOR);
        g.setStroke(new BasicStroke(10));
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g.drawOval((WIDTH * xCord) / 100, (HEIGHT * yCord) / 100, MARKER_WIDTH, MARKER_WIDTH);

        ImageIO.write(image, "png", path.toFile());
    }

    /**
     * Stáhne obrázek do dané lokace.
     *
     * @param location cesta k uložení obrázku.
     * @param urlToDownload URL ke stažení obrázku.
     * */
    public void download(@NotNull final Path location, @NotNull String urlToDownload) throws IOException {
        FileUtils.copyURLToFile(new URL(urlToDownload), location.toFile(), 50000, 500000);
    }

    /**
     * Vytvoří API pro tile server.
     *
     * @param latitude zeměpisná šířka
     * @param longitude zeměpisná délka
     * @param zoom zoom.
     * */
    protected TileInfo createAPIURL(@NotNull Double latitude, @NotNull final Double longitude, @NotNull final Integer zoom){
        String main = environment.getProperty("openstreetmap.tile.main");
        TileInfo info = getTileNumber(latitude, longitude, zoom);
        String url = String.format("%s%d/%d/%d.png", main, zoom, info.getxTile(), info.getyTile());

        info.setUrl(url);

        return info;
    }

    /**
     * https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
     */
    public TileInfo getTileNumber(final double lat, final double lon, final int zoom) {
        double x = (lon + 180) / 360 * (1 << zoom);
        double y = (1 - Math.log(Math.tan(Math.toRadians(lat)) + 1 / Math.cos(Math.toRadians(lat))) / Math.PI) / 2 * (1 << zoom);

        int xtile = (int)  Math.floor(x);
        int ytile = (int)  Math.floor(y);

        BigDecimal xbig = new BigDecimal(String.valueOf(x));
        int xCord = getDecimalPart(xbig);

        BigDecimal ybig = new BigDecimal(String.valueOf(y));
        int yCord = getDecimalPart(ybig);

        if (xtile < 0)
            xtile = 0;
        if (xtile >= (1 << zoom))
            xtile = ((1 << zoom) - 1);
        if (ytile < 0)
            ytile = 0;
        if (ytile >= (1 << zoom))
            ytile = ((1 << zoom) - 1);
        return new TileInfo(xCord, yCord, xtile, ytile);
    }

    /**
     * Vrátí desetinou část čísla.
     * @param decimal desetinné číslo.
     * */
    public Integer getDecimalPart(final BigDecimal decimal){
        String[] parts = decimal.toPlainString().split("\\.");

        if(parts.length < 2)
            return 0;
        if(parts[1].length() == 1)
            return Integer.parseInt(parts[1]);
        else
            return Integer.parseInt(parts[1].substring(0, 2));
    }

    /**
     * Jednoduchá třída pro reprezentaci výsledků.
     *
     * @author Ondřej Chrbolka
     * @version 1.0
     *
     * */
    private static class TileInfo{
        private String url;
        private final Integer xCord;
        private final Integer yCord;
        private final Integer xTile;
        private final Integer yTile;

        public TileInfo(Integer xCord, Integer yCord, Integer xTile, Integer yTile) {
            this.xCord = xCord;
            this.yCord = yCord;
            this.xTile = xTile;
            this.yTile = yTile;
        }

        public String getUrl() {
            return url;
        }

        public Integer getxCord() {
            return xCord;
        }

        public Integer getyCord() {
            return yCord;
        }

        public Integer getxTile() {
            return xTile;
        }

        public Integer getyTile() {
            return yTile;
        }

        public void setUrl(String url) {
            this.url = url;
        }
    }
}
